package dev.fitko.fitconnect.client.cli;

import dev.fitko.fitconnect.api.domain.model.metadata.attachment.Attachment;
import dev.fitko.fitconnect.api.domain.model.metadata.attachment.AttachmentWithData;
import dev.fitko.fitconnect.api.domain.model.metadata.data.MimeType;
import dev.fitko.fitconnect.api.domain.model.submission.ReceivedSubmission;
import dev.fitko.fitconnect.api.domain.model.submission.SentSubmission;
import dev.fitko.fitconnect.api.domain.model.submission.SubmissionForPickup;
import dev.fitko.fitconnect.client.SenderClient;
import dev.fitko.fitconnect.client.SubscriberClient;
import dev.fitko.fitconnect.client.cli.batch.CsvImporter;
import dev.fitko.fitconnect.client.cli.batch.ImportRecord;
import dev.fitko.fitconnect.client.testutil.LogCaptor;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.api.io.TempDir;

import java.nio.file.Path;
import java.util.*;

import static org.mockito.ArgumentMatchers.any;
import static org.mockito.Mockito.mock;
import static org.mockito.Mockito.when;

class CommandLineClientTest {

    private final LogCaptor logs = new LogCaptor();

    private CommandLineClient underTest;
    private SenderClient.ClientBuilder senderClientMock;
    private SubscriberClient.ClientBuilder subscriberClientMock;

    @BeforeEach
    void setUp() {
        this.senderClientMock = mock(SenderClient.ClientBuilder.class);
        this.subscriberClientMock = mock(SubscriberClient.ClientBuilder.class);
        this.underTest = new CommandLineClient(this.senderClientMock, this.subscriberClientMock);
    }

    @Test
    void testInvalidDestinationId() {
        // When
        this.underTest.run("list", "--destinationId=123");

        // Then
        this.logs.assertContains("Parameter 123 is not a valid UUID.");
    }

    @Test
    void testListAllSubmissions() {

        // Given
        final var destinationId = UUID.randomUUID();
        final Set<SubmissionForPickup> submissions = generateSubmissions(destinationId, 10);
        when(this.subscriberClientMock.getAvailableSubmissions(destinationId)).thenReturn(submissions);

        // When
        this.underTest.run("list", "--destinationId=" + destinationId);

        // Then
        this.logs.assertContains("Listing available submissions for destination " + destinationId);
        submissions.forEach(submission -> this.logs.assertContains(submission.getSubmissionId().toString()));
    }

    @Test
    void testGetSingleSubmission(@TempDir final Path tempDir) {

        // Given
        final var submissionId = UUID.randomUUID();

        final var attachmentMetadata = new Attachment();
        attachmentMetadata.setFilename("testAttachment.txt");
        final AttachmentWithData attachment = AttachmentWithData.builder()
                .decryptedData("attachment text".getBytes())
                .attachmentMetadata(attachmentMetadata)
                .build();

        final var receivedSubmission = ReceivedSubmission.builder()
                .data("{ data: '' }")
                .attachments(List.of(attachment))
                .dataMimeType(MimeType.APPLICATION_JSON)
                .build();

        when(this.subscriberClientMock.requestSubmission(submissionId)).thenReturn(this.subscriberClientMock);
        when(this.subscriberClientMock.getSubmissionData()).thenReturn(Optional.of(receivedSubmission));

        // When
        this.underTest.run("get", "--submissionId=" + submissionId, "--target=" + tempDir);

        // Then
        this.logs.assertContains("Getting submission for id " + submissionId);
        this.logs.assertContains("Creating data directory for submission in " + tempDir);
        this.logs.assertContains("Writing data.json");
        this.logs.assertContains("Writing attachment testAttachment.txt");
    }

    @Test
    void testSendSubmission() {

        // Given
        final var destinationId = UUID.randomUUID();

        final var expectedSubmission = SentSubmission.builder()
                .submissionId(UUID.randomUUID())
                .build();

        when(this.senderClientMock.withDestination(any())).thenReturn(this.senderClientMock);
        when(this.senderClientMock.withServiceType(any(), any())).thenReturn(this.senderClientMock);
        when(this.senderClientMock.withAttachments(any())).thenReturn(this.senderClientMock);
        when(this.senderClientMock.withJsonData(any())).thenReturn(this.senderClientMock);
        when(this.senderClientMock.submit()).thenReturn(Optional.of(expectedSubmission));
        when(this.senderClientMock.newSubmission()).thenReturn(this.senderClientMock);

        // When
        this.underTest.run("send",
                "--destinationId=" + destinationId,
                "--serviceName=Test",
                "--leikaKey=urn:de:fim:leika:leistung:99400048079000",
                "--attachments=src/test/resources/attachment.txt",
                "--data=src/test/resources/data.json",
                "--dataType=json");

        // Then
        this.logs.assertContains("Sending new submission to destination " + destinationId);
    }

    @Test
    void testBatchImport() {

        // Given

        final var testDataPath = "src/test/resources/batch_data.csv";
        final List<ImportRecord> importRecords = new CsvImporter().readRecords(testDataPath);

        when(this.senderClientMock.withDestination(any())).thenReturn(this.senderClientMock);
        when(this.senderClientMock.withServiceType(any(), any())).thenReturn(this.senderClientMock);
        when(this.senderClientMock.withAttachments(any())).thenReturn(this.senderClientMock);
        when(this.senderClientMock.withJsonData(any())).thenReturn(this.senderClientMock);
        when(this.senderClientMock.submit()).thenReturn(Optional.of(SentSubmission.builder().build()));
        when(this.senderClientMock.newSubmission()).thenReturn(this.senderClientMock);

        // When
        this.underTest.run("batch", "--data=" + testDataPath);

        // Then
        this.logs.assertContains("DONE ! Finished batch import of " + importRecords.size() + " submissions");
    }

    private Set<SubmissionForPickup> generateSubmissions(final UUID destinationId, final int count) {
        final Set<SubmissionForPickup> submissions = new HashSet<>(count);
        for (int i = 0; i < count; i++) {
            final SubmissionForPickup submission = new SubmissionForPickup();
            submission.setSubmissionId(UUID.randomUUID());
            submission.setDestinationId(destinationId);
            submissions.add(submission);
        }
        return submissions;
    }
}